@@ -23,9 +23,6 @@ use log::{debug, trace, warn};
 #[cfg(feature = "defmt")]
 use defmt::{debug, trace, warn};
 
-extern crate modular_bitfield;
-use modular_bitfield::prelude::*;
-
 #[cfg(feature = "serde")]
 extern crate serde;
 
@@ -37,7 +34,9 @@ use hal::blocking::spi::{Transfer, Write};
 use hal::digital::v2::OutputPin;
 
 extern crate radio;
-use radio::{Channel as _, Interrupts as _, Power as _, Rssi as _, State as _};
+use radio::{
+    Channel as _, Interrupts as _, Power as _, Register as _, Registers as _, Rssi as _, State as _,
+};
 
 pub mod prelude;
 use prelude::*;
@@ -91,12 +90,16 @@ pub enum Error<SpiError, PinError> {
     InvalidDevice(u8),
     /// Packet size disagrees with FIFO interrupt (first is packet len, second is buffer size)
     InvalidPacketSize(usize, usize),
+    // #[error("Unexpected register value (address: 0x{0:02x})")]
+    UnexpectedValue(u8),
 }
 
 impl<Spi, CsPin, SpiError, PinError> Sx1231<Spi, CsPin, SpiError, PinError>
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     /// Create a new radio instance with the provided SPI implementation and pins
     pub fn new(spi: Spi, cs: CsPin) -> Self {
@@ -121,6 +124,8 @@ impl<Spi, CsPin, SpiError, PinError> Sx1231<Spi, CsPin, SpiError, PinError>
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     /// (re)apply device configuration
     pub fn configure(&mut self, config: &Config) -> Result<(), Error<SpiError, PinError>> {
@@ -133,112 +138,79 @@ where
         // // Set channel configuration
         self.set_channel(&config.channel)?;
 
-        self.write_regs(
-            DataModulation::ADDRESS,
-            &DataModulation::new()
+        self.write_register(
+            DataModulation::new()
                 .with_mod_type(config.modulation)
-                .with_mode(config.data_mode)
-                .into_bytes(),
+                .with_mode(config.data_mode),
         )?;
 
-        // // Set preamble length
-        let preamble_len = [
-            (config.preamble_len >> 8) as u8,
-            (config.preamble_len & 0xFF) as u8,
-        ];
-        self.write_regs(Preamble::ADDRESS, &preamble_len)?;
+        self.write_register(Preamble::new().with_length(config.preamble_len))?;
 
         self.set_sync_word(config.sync_word)?;
 
         // Set payload length
         let packet_format = match config.payload_mode {
             PayloadMode::Constant(v) => {
-                self.write_reg(PayloadLength::ADDRESS, v as u8)?;
+                self.write_register(PayloadLength::new().with_length(v as u8))?;
                 PacketFormat::Fixed
             }
             PayloadMode::Variable => {
-                self.write_reg(PayloadLength::ADDRESS, 0xFF)?;
+                self.write_register(PayloadLength::new().with_length(0xFF))?;
                 PacketFormat::Variable
             }
         };
 
         // Set packetconfig1 register
-        self.write_regs(
-            PacketConfig1::ADDRESS,
-            &PacketConfig1::new()
+        self.write_register(
+            PacketConfig1::new()
                 .with_packet_format(packet_format)
                 .with_dc_free(config.dc_free)
                 .with_crc(config.crc)
                 .with_crc_auto_clear(config.crc_auto_clear)
-                .with_address_filter(config.address_filter)
-                .into_bytes(),
+                .with_address_filter(config.address_filter),
         )?;
 
         // Set packetconfig2 register
-        self.write_regs(
-            PacketConfig2::ADDRESS,
-            &PacketConfig2::new()
+        self.write_register(
+            PacketConfig2::new()
                 .with_auto_rx_restart(config.auto_rx_restart)
-                .with_inter_packet_rx_delay(config.inter_packet_rx_delay)
-                .into_bytes(),
+                .with_inter_packet_rx_delay(config.inter_packet_rx_delay),
         )?;
 
         // Configure TXStart
-        self.write_regs(
-            FifoThreshold::ADDRESS,
-            &FifoThreshold::new()
+        self.write_register(
+            FifoThreshold::new()
                 .with_start_condition(config.tx_start_condition)
-                .with_threshold(config.fifo_threshold)
-                .into_bytes(),
+                .with_threshold(config.fifo_threshold),
         )?;
 
-        self.write_regs(
-            RssiThreshold::ADDRESS,
-            &RssiThreshold::new()
-                .with_value(config.rssi_threshold)
-                .into_bytes(),
-        )?;
+        self.write_register(RssiThreshold::new().with_value(config.rssi_threshold))?;
 
-        self.write_regs(
-            AfcControl::ADDRESS,
-            &AfcControl::new()
-                .with_low_beta_on(config.afc_low_beta)
-                .into_bytes(),
-        )?;
+        self.write_register(AfcControl::new().with_low_beta_on(config.afc_low_beta))?;
 
         if config.afc_low_beta {
-            self.write_regs(
-                0x6f, // RegTestDagc
-                &[0x20],
-            )?; // Improved margin, use if AfcLowBetaOn=1
-            self.write_regs(
-                0x71, // RegTestAfc
-                &[20],
-            )?; // LowBetaAfcOffset = 20 * 488 = 9.8kHz (should be ~10% of RxBw and > DCC cutoff)
+            // Improved margin, use if AfcLowBetaOn=1
+            self.write_register(
+                TestDagc::new().with_continous_dagc(ContinousDagc::ImprovedMarginLowBetaOn),
+            )?;
+            // LowBetaAfcOffset = 20 * 488 = 9.8kHz (should be ~10% of RxBw and > DCC cutoff)
+            self.write_register(TestAfc::new().with_low_beta_offset(20))?;
         }
 
-        self.write_regs(
-            AfcFei::ADDRESS,
-            &AfcFei::new().with_afc_auto_on(true).into_bytes(),
-        )?;
+        self.write_register(AfcFei::new().with_afc_auto_on(true))?;
 
-        self.write_regs(
-            0x2bu8, //timeout rssi threshold
-            &[64u8],
-        )?; // 64 * 16 * 10 us = 12ms
+        // 64 * 16 * 10 us = 12ms
+        self.write_register(TimeoutRssiThresh::new().with_value(64))?;
 
-        self.write_regs(
-            Lna::ADDRESS,
-            &Lna::new()
+        self.write_register(
+            Lna::new()
                 .with_zin(LnaZin::Zin50Ohm)
-                .with_gain_select(LnaGainSelect::Agc)
-                .into_bytes(),
+                .with_gain_select(LnaGainSelect::Agc),
         )?;
 
-        self.write_regs(
-            0x58, // RegTestLna
-            &[0x2d],
-        )?; // SensitivityBoost
+        self.write_register(
+            TestLna::new().with_sensitivity_boost(SensitivityBoost::HighSensitivity),
+        )?;
 
         // Configure power amplifier
         self.set_power(config.power)?;
@@ -251,26 +223,23 @@ where
 
     fn set_sync_word(&mut self, sync_word: &[u8]) -> Result<(), Error<SpiError, PinError>> {
         if !sync_word.is_empty() {
-            self.write_regs(
-                SyncConfig::ADDRESS,
-                &SyncConfig::new()
+            self.write_register(
+                SyncConfig::new()
                     .with_sync_on(true)
-                    .with_size_minus_one(sync_word.len() as u8 - 1)
-                    .into_bytes(),
+                    .with_size_minus_one(sync_word.len() as u8 - 1),
             )?;
-            self.write_regs(SyncValue::ADDRESS, sync_word)?;
+            let mut sync_value = [0u8; 8];
+            sync_value[..sync_word.len()].copy_from_slice(sync_word);
+            self.write_register(SyncValue::new().with_value(u64::from_be_bytes(sync_value)))?;
         } else {
-            self.write_regs(
-                SyncConfig::ADDRESS,
-                &SyncConfig::new().with_sync_on(false).into_bytes(),
-            )?;
+            self.write_register(SyncConfig::new().with_sync_on(false))?;
         }
         Ok(())
     }
 
     /// Fetch device silicon version
     pub fn silicon_version(&mut self) -> Result<u8, Error<SpiError, PinError>> {
-        self.read_reg(Version::ADDRESS)
+        self.read_register::<Version>().map(|r| r.into())
     }
 
     // Calculate a channel number from a given frequency
@@ -283,103 +252,11 @@ where
     // Set the channel by frequency
     pub fn set_frequency(&mut self, freq: u32) -> Result<(), Error<SpiError, PinError>> {
         let channel = self.freq_to_channel_index(freq);
-
-        let outgoing = [(channel >> 16) as u8, (channel >> 8) as u8, channel as u8];
-
         debug!("Set channel to index: {:?} (freq: {:?})", channel, freq);
-
-        self.write_regs(CarrierFreq::ADDRESS, &outgoing)?;
-
-        Ok(())
+        self.write_register(CarrierFreq::new().with_freq(channel))
     }
 
-    /// Read a u8 value from the specified register
-    pub fn read_reg<R>(&mut self, reg: R) -> Result<u8, Error<SpiError, PinError>>
-    where
-        R: Copy + Clone + Debug + Into<u8>,
-    {
-        let mut incoming = [0u8; 1];
-        self.read_regs(reg.into(), &mut incoming)?;
-        Ok(incoming[0])
-    }
-
-    pub fn read_register<R>(&mut self) -> Result<R::InOut, Error<SpiError, PinError>>
-    where
-        R: Reg + Specifier<Bytes = u8>,
-    {
-        let data = self.read_reg(R::ADDRESS)?;
-        let reg = R::from_bytes(data).unwrap();
-        Ok(reg)
-    }
-
-    /// Write a u8 value to the specified register
-    pub fn write_reg<R>(&mut self, reg: R, value: u8) -> Result<(), Error<SpiError, PinError>>
-    where
-        R: Copy + Clone + Debug + Into<u8>,
-    {
-        self.write_regs(reg.into(), &[value])?;
-
-        Ok(())
-    }
-
-    /// Update the specified register with the provided value & mask
-    pub fn update_reg<R>(
-        &mut self,
-        reg: R,
-        mask: u8,
-        value: u8,
-    ) -> Result<(), Error<SpiError, PinError>>
-    where
-        R: Copy + Clone + Debug + Into<u8>,
-    {
-        let existing = self.read_reg(reg)?;
-        let updated = (existing & !mask) | (value & mask);
-        self.write_reg(reg, updated)?;
-
-        Ok(())
-    }
-
-    /// Read from the specified register
-    fn read_regs<R>(&mut self, reg: R, data: &mut [u8]) -> Result<(), Error<SpiError, PinError>>
-    where
-        R: Copy + Clone + Debug + Into<u8>,
-    {
-        self.cs.set_low().map_err(|e| Error::Pin(e))?;
-
-        // Send command then read data
-        self.spi
-            .write(&[reg.into() & 0x7F])
-            .map_err(|e| Error::Spi(e))?;
-        self.spi.transfer(data).map_err(|e| Error::Spi(e))?;
-
-        self.cs.set_high().map_err(|e| Error::Pin(e))?;
-
-        Ok(())
-    }
-
-    /// Write to the specified register
-    fn write_regs<R>(&mut self, reg: R, data: &[u8]) -> Result<(), Error<SpiError, PinError>>
-    where
-        R: Copy + Clone + Debug + Into<u8>,
-    {
-        self.cs.set_low().map_err(|e| Error::Pin(e))?;
-        #[cfg(feature = "log")]
-        trace!("Write reg:  {:?} (0x{:02x}): {:x?}", reg, reg.into(), data);
-        #[cfg(feature = "defmt")]
-        trace!("Write reg:  {} (0x{}): {}", reg, reg.into(), data);
-
-        // Send command then write data
-        self.spi
-            .write(&[reg.into() | 0x80])
-            .map_err(|e| Error::Spi(e))?;
-        self.spi.write(data).map_err(|e| Error::Spi(e))?;
-
-        self.cs.set_high().map_err(|e| Error::Pin(e))?;
-
-        Ok(())
-    }
-
-    /// Write to the specified buffer
+    /// Write to the FIFO buffer
     fn write_fifo(&mut self, data: &[u8]) -> Result<(), Error<SpiError, PinError>> {
         self.cs.set_low().map_err(|e| Error::Pin(e))?;
 
@@ -396,7 +273,7 @@ where
         Ok(())
     }
 
-    /// Read from the specified buffer
+    /// Read from the FIFO buffer
     fn read_fifo(&mut self, len: usize) -> Result<(), Error<SpiError, PinError>> {
         let rx_buf_len = self.rx_buf_len;
         if rx_buf_len + len > self.rx_buf.len() {
@@ -429,6 +306,8 @@ impl<Spi, CsPin, SpiError, PinError> Sx1231<Spi, CsPin, SpiError, PinError>
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     pub fn set_state_checked(&mut self, state: ModemMode) -> Result<(), Error<SpiError, PinError>> {
         // Send set state command
@@ -460,22 +339,22 @@ impl<Spi, CsPin, SpiError, PinError> radio::State for Sx1231<Spi, CsPin, SpiErro
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type State = ModemMode;
     type Error = Error<SpiError, PinError>;
 
     /// Fetch device modem mode
     fn get_state(&mut self) -> Result<Self::State, Self::Error> {
-        let op_mode = OpMode::from_bytes([self.read_reg(OpMode::ADDRESS)?]);
-        Ok(op_mode.modem_mode())
+        Ok(self.read_register::<OpMode>()?.modem_mode())
     }
 
     /// Set device modem mode
     fn set_state(&mut self, state: Self::State) -> Result<(), Self::Error> {
         let mut op_mode = self.read_register::<OpMode>()?;
         op_mode.set_modem_mode(state);
-        self.write_regs(OpMode::ADDRESS, &op_mode.into_bytes())?;
-        Ok(())
+        self.write_register(op_mode)
     }
 }
 
@@ -483,6 +362,8 @@ impl<Spi, CsPin, SpiError, PinError> radio::Power for Sx1231<Spi, CsPin, SpiErro
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Error = Error<SpiError, PinError>;
 
@@ -505,7 +386,7 @@ where
 
         debug!("Updated PA config for {} dBm: {:?}", power, config);
 
-        self.write_regs(PaLevel::ADDRESS, &config.into_bytes())?;
+        self.write_register(config)?;
 
         Ok(())
     }
@@ -515,6 +396,8 @@ impl<Spi, CsPin, SpiError, PinError> radio::Interrupts for Sx1231<Spi, CsPin, Sp
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Irq = IrqFlags;
     type Error = Error<SpiError, PinError>;
@@ -522,15 +405,10 @@ where
     /// Fetch pending interrupts from the device
     /// If the clear option is set, this will also clear any pending flags
     fn get_interrupts(&mut self, clear: bool) -> Result<Self::Irq, Self::Error> {
-        let mut regs = [0u8; 2];
-        self.read_regs(IrqFlags::ADDRESS, &mut regs)?;
-        let irq = IrqFlags::from_bytes(regs);
+        let irq: IrqFlags = self.read_register()?;
 
         if clear {
-            self.write_regs(
-                IrqFlags::ADDRESS,
-                &irq.with_rssi(false).with_fifo_overrun(false).into_bytes(),
-            )?;
+            self.write_register(irq.with_rssi(false).with_fifo_overrun(false))?;
         }
 
         Ok(irq)
@@ -541,6 +419,8 @@ impl<Spi, CsPin, SpiError, PinError> radio::Channel for Sx1231<Spi, CsPin, SpiEr
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Channel = Channel;
     type Error = Error<SpiError, PinError>;
@@ -567,19 +447,15 @@ where
         trace!("fdev: {} bitrate: {}", fdev, datarate);
 
         // Set frequency deviation
-        self.write_regs(FreqDev::ADDRESS, &[(fdev >> 8) as u8, (fdev & 0xFF) as u8])?;
-        // self.write_regs(FreqDev::ADDRESS, FreqDev::new().with_freq_dev(fdev).into_bytes())?;
+        self.write_register(FreqDev::new().with_freq_dev(fdev))?;
 
         // Set bitrate
         // self.write_regs(Bitrate::ADDRESS, Bitrate::new().with_bitrate(datarate).into_bytes())?;
-        self.write_regs(
-            Bitrate::ADDRESS,
-            &[(datarate >> 8) as u8, (datarate & 0xFF) as u8],
-        )?;
+        self.write_register(Bitrate::new().with_bitrate(datarate))?;
 
         // Set bandwidths
-        self.write_reg(RxBw::ADDRESS, 0b0100_0000 | channel.bw as u8)?;
-        self.write_reg(AfcBw::ADDRESS, 0b0100_0000 | channel.bw_afc as u8)?;
+        self.write_register(RxBw::new().with_bw(channel.bw))?;
+        self.write_register(AfcBw::new().with_bw(channel.bw_afc))?;
 
         Ok(())
     }
@@ -589,6 +465,8 @@ impl<Spi, CsPin, SpiError, PinError> radio::Transmit for Sx1231<Spi, CsPin, SpiE
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Error = Error<SpiError, PinError>;
 
@@ -656,6 +534,8 @@ impl<Spi, CsPin, SpiError, PinError> radio::Receive for Sx1231<Spi, CsPin, SpiEr
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Info = PacketInfo;
     type Error = Error<SpiError, PinError>;
@@ -741,11 +621,8 @@ where
     ///
     /// This copies data into the provided slice, updates the provided information object,
     ///  and returns the number of bytes received on success
-    fn get_received(
-        &mut self,
-        info: &mut Self::Info,
-        data: &mut [u8],
-    ) -> Result<usize, Self::Error> {
+    fn get_received(&mut self, data: &mut [u8]) -> Result<(usize, Self::Info), Self::Error> {
+        let mut info = PacketInfo::default();
         // Read the RSSI
         info.rssi = self.poll_rssi()?;
         // First byte is length and should not be returned.
@@ -758,7 +635,7 @@ where
         #[cfg(feature = "defmt")]
         debug!("Received data: {} info: {:?}", packet_data, &info);
 
-        Ok(packet_data.len())
+        Ok((packet_data.len(), info))
     }
 }
 
@@ -766,18 +643,116 @@ impl<Spi, CsPin, SpiError, PinError> radio::Rssi for Sx1231<Spi, CsPin, SpiError
 where
     Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
     CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
 {
     type Error = Error<SpiError, PinError>;
 
     /// Poll for the current channel RSSI
     /// This should only be called in receive mode
     fn poll_rssi(&mut self) -> Result<i16, Error<SpiError, PinError>> {
-        let rssi = -(self.read_reg(RssiValue::ADDRESS)? as i16) / 2;
-
+        let reg: u8 = self.read_register::<RssiValue>()?.into();
+        let rssi = -(i16::from(reg)) / 2;
         Ok(rssi)
     }
 }
 
+pub trait RegisterWord {
+    type Bytes: Default + core::convert::AsMut<[u8]>;
+    fn to_bytes(self) -> Self::Bytes;
+    fn from_bytes(bytes: Self::Bytes) -> Self;
+}
+
+impl RegisterWord for u8 {
+    type Bytes = [u8; 1];
+
+    fn to_bytes(self) -> Self::Bytes {
+        self.to_be_bytes()
+    }
+
+    fn from_bytes(bytes: Self::Bytes) -> Self {
+        Self::from_be_bytes(bytes)
+    }
+}
+
+impl RegisterWord for u16 {
+    type Bytes = [u8; 2];
+
+    fn to_bytes(self) -> Self::Bytes {
+        self.to_be_bytes()
+    }
+
+    fn from_bytes(bytes: Self::Bytes) -> Self {
+        Self::from_be_bytes(bytes)
+    }
+}
+
+impl RegisterWord for [u8; 3] {
+    type Bytes = [u8; 3];
+
+    fn to_bytes(self) -> Self::Bytes {
+        self
+    }
+
+    fn from_bytes(bytes: Self::Bytes) -> Self {
+        bytes
+    }
+}
+
+impl RegisterWord for u64 {
+    type Bytes = [u8; 8];
+
+    fn to_bytes(self) -> Self::Bytes {
+        self.to_be_bytes()
+    }
+
+    fn from_bytes(bytes: Self::Bytes) -> Self {
+        Self::from_be_bytes(bytes)
+    }
+}
+
+impl<Spi, CsPin, SpiError, PinError, Word, Bytes> radio::Registers<Word>
+    for Sx1231<Spi, CsPin, SpiError, PinError>
+where
+    Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError>,
+    CsPin: OutputPin<Error = PinError>,
+    SpiError: Debug,
+    PinError: Debug,
+    Word: RegisterWord<Bytes = Bytes>,
+    Bytes: Default + core::convert::AsMut<[u8]>,
+{
+    type Error = Error<SpiError, PinError>;
+
+    fn read_register<R: radio::Register<Word = Word>>(
+        &mut self,
+    ) -> Result<R, Error<SpiError, PinError>> {
+        let mut bytes = Bytes::default();
+
+        self.cs.set_low().map_err(|e| Error::Pin(e))?;
+        self.spi.write(&[R::ADDRESS]).map_err(|e| Error::Spi(e))?;
+        self.spi
+            .transfer(bytes.as_mut())
+            .map_err(|e| Error::Spi(e))?;
+        self.cs.set_high().map_err(|e| Error::Pin(e))?;
+
+        let d = R::Word::from_bytes(bytes);
+        R::try_from(d).map_err(|_| Error::UnexpectedValue(R::ADDRESS))
+    }
+
+    fn write_register<R: radio::Register<Word = Word>>(
+        &mut self,
+        r: R,
+    ) -> Result<(), Error<SpiError, PinError>> {
+        self.cs.set_low().map_err(|e| Error::Pin(e))?;
+        self.spi.write(&[R::ADDRESS]).map_err(|e| Error::Spi(e))?;
+        self.spi
+            .write(r.into().to_bytes().as_mut())
+            .map_err(|e| Error::Spi(e))?;
+        self.cs.set_high().map_err(|e| Error::Pin(e))?;
+        Ok(())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     #[test]
